@@ -70,11 +70,9 @@ from .parse_utils import (
 from .root import Root
 from .utils import (
     AsyncExitStack,
-    alias,
     argument,
     calc_life_span,
     command,
-    deprecated_quiet_option,
     group,
     option,
     resolve_job,
@@ -114,29 +112,13 @@ def job() -> None:
 
 @command(context_settings=dict(allow_interspersed_args=False))
 @argument("job", type=JOB)
-@argument("cmd", nargs=-1, type=click.UNPROCESSED, required=True)
+@argument("cmd", nargs=-1, type=click.UNPROCESSED, metavar="-- CMD...", required=True)
 @TTY_OPT
-@option(
-    "--no-key-check",
-    is_flag=True,
-    help="Disable host key checks. Should be used with caution.",
-    hidden=True,
-)
-@option(
-    "--timeout",
-    default=0,
-    type=float,
-    show_default=True,
-    hidden=True,
-    help="Maximum allowed time for executing the command, 0 for no timeout",
-)
 async def exec(
     root: Root,
     job: str,
     tty: Optional[bool],
-    no_key_check: bool,
     cmd: Sequence[str],
-    timeout: float,
 ) -> None:
     """
     Execute command in a running job.
@@ -144,11 +126,12 @@ async def exec(
     Examples:
 
     # Provides a shell to the container:
-    neuro exec my-job /bin/bash
+    neuro exec my-job -- /bin/bash
 
     # Executes a single command in the container and returns the control:
-    neuro exec --no-tty my-job ls -l
+    neuro exec --no-tty my-job -- ls -l
     """
+    cmd = _fix_cmd("neuro exec", "JOB -- CMD...", cmd)
     real_cmd = _parse_cmd(cmd)
     job, cluster_name = await resolve_job_ex(
         job,
@@ -170,14 +153,8 @@ async def exec(
     required=True,
     metavar="LOCAL_PORT:REMOTE_RORT...",
 )
-@option(
-    "--no-key-check",
-    is_flag=True,
-    help="Disable host key checks. Should be used with caution.",
-    hidden=True,
-)
 async def port_forward(
-    root: Root, job: str, no_key_check: bool, local_remote_port: List[Tuple[int, int]]
+    root: Root, job: str, local_remote_port: List[Tuple[int, int]]
 ) -> None:
     """
     Forward port(s) of a running job to local port(s).
@@ -198,13 +175,6 @@ async def port_forward(
     neuro job port-forward my-job 2080:80 2222:22 2000:100
 
     """
-    if no_key_check:
-        click.secho(
-            "--no-key-check option is deprecated and "
-            "will be removed from the future Neuro CLI release",
-            fg="red",
-            err=True,
-        )
     job_id, cluster_name = await resolve_job_ex(
         job,
         client=root.client,
@@ -232,7 +202,19 @@ async def port_forward(
 
 @command()
 @argument("job", type=JOB)
-async def logs(root: Root, job: str) -> None:
+@option(
+    "--since",
+    metavar="DATE_OR_TIMEDELTA",
+    help="Only return logs after a specific date (including). "
+    "Use value of format '1d2h3m4s' to specify moment in "
+    "past relatively to current time.",
+)
+@option(
+    "--timestamps",
+    is_flag=True,
+    help="Include timestamps on each line in the log output.",
+)
+async def logs(root: Root, since: str, job: str, timestamps: bool) -> None:
     """
     Print the logs for a job.
     """
@@ -241,7 +223,14 @@ async def logs(root: Root, job: str) -> None:
         client=root.client,
         status=JobStatus.items(),
     )
-    await process_logs(root, id, None, cluster_name=cluster_name)
+    await process_logs(
+        root,
+        id,
+        None,
+        cluster_name=cluster_name,
+        since=_parse_date(since),
+        timestamps=timestamps,
+    )
 
 
 @command()
@@ -352,7 +341,6 @@ async def attach(root: Root, job: str, port_forward: List[Tuple[int, int]]) -> N
     default=False,
     help="Show all jobs regardless the status.",
 )
-@deprecated_quiet_option
 @option("-w", "--wide", is_flag=True, help="Do not cut long lines for terminal width.")
 @option(
     "--format",
@@ -772,7 +760,7 @@ async def top(
         await asyncio.gather(create_pollers(), renderer())
 
 
-@command()
+@command(deprecated=True, hidden=True)
 @argument("job", type=JOB)
 @argument("image", type=RemoteImageType())
 async def save(root: Root, job: str, image: RemoteImage) -> None:
@@ -825,7 +813,7 @@ async def kill(root: Root, jobs: Sequence[str]) -> None:
 
 @command(context_settings=dict(allow_interspersed_args=False))
 @argument("image", type=RemoteImageType())
-@argument("cmd", nargs=-1, type=click.UNPROCESSED)
+@argument("cmd", nargs=-1, type=click.UNPROCESSED, metavar="[-- CMD...]")
 @option(
     "--cluster",
     type=CLUSTER,
@@ -864,13 +852,6 @@ async def kill(root: Root, jobs: Sequence[str]) -> None:
     help="Enable HTTP authentication for forwarded HTTP port  [default: True]",
     default=None,
 )
-@option(
-    "--preemptible/--non-preemptible",
-    "-p/-P",
-    help="Run job on a lower-cost preemptible instance (DEPRECATED AND IGNORED)",
-    default=None,
-    hidden=True,
-)
 @option(
     "-n",
     "--name",
@@ -894,7 +875,6 @@ async def kill(root: Root, jobs: Sequence[str]) -> None:
     help="Optional job description in free format",
     secure=True,
 )
-@deprecated_quiet_option
 @option(
     "-v",
     "--volume",
@@ -1035,7 +1015,6 @@ async def run(
     env_file: Sequence[str],
     restart: str,
     life_span: Optional[str],
-    preemptible: Optional[bool],
     name: Optional[str],
     tag: Sequence[str],
     description: Optional[str],
@@ -1068,17 +1047,14 @@ async def run(
 
     # Starts a container using the custom image my-ubuntu:latest stored in neuro
     # registry, run /script.sh and pass arg1 and arg2 as its arguments:
-    neuro run -s cpu-small image:my-ubuntu:latest --entrypoint=/script.sh arg1 arg2
+    neuro run -s cpu-small --entrypoint=/script.sh image:my-ubuntu:latest -- arg1 arg2
     """
+    cmd = _fix_cmd("neuro run", "IMAGE -- CMD...", cmd)
     cluster_name = cluster or root.client.cluster_name
     cluster_config = root.client.config.clusters[cluster_name]
     if not preset:
         preset = next(iter(cluster_config.presets.keys()))
     job_preset = cluster_config.presets[preset]
-    if preemptible is not None:
-        root.print(
-            "-p/-P option is deprecated and ignored. Use corresponding presets instead."
-        )
     log.info(f"Using preset '{preset}': {job_preset}")
     if tty is None:
         tty = root.tty
@@ -1151,10 +1127,6 @@ job.add_command(attach)
 job.add_command(bump_life_span)
 
 
-job.add_command(alias(ls, "list", hidden=True))
-job.add_command(alias(logs, "monitor", hidden=True))
-
-
 async def run_job(
     root: Root,
     *,
@@ -1456,3 +1428,17 @@ def _job_to_cli_args(job: JobDescription) -> List[str]:
     if job.container.command:
         res += [job.container.command]
     return res
+
+
+def _fix_cmd(cmd_name: str, tmpl: str, cmd: Sequence[str]) -> Sequence[str]:
+    if cmd:
+        if cmd[0] != "--":
+            old_pre, sep, old_suff = tmpl.partition("--")
+            old = old_pre.strip() + " " + old_suff.strip()
+            log.warning(
+                f"'{cmd_name} {old}' is DEPRECATED and will be removed "
+                f"in a future version.\nUse '{cmd_name} {tmpl}' instead."
+            )
+        else:
+            return cmd[1:]
+    return cmd
