@@ -22,14 +22,12 @@ from typing import (
     Tuple,
     Type,
     TypeVar,
-    Union,
     cast,
 )
 
 import click
 import humanize
 from aiohttp import ClientResponseError
-from click.types import convert_type
 from yarl import URL
 
 from neuro_sdk import Action, Client, JobStatus, Volume
@@ -400,18 +398,11 @@ class Option(click.Option):
 def option(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
     option_attrs = attrs.copy()
     option_attrs.setdefault("cls", Option)
-    typ = convert_type(attrs.get("type"), attrs.get("default"))
-    autocompletion = getattr(typ, "complete", None)
-    option_attrs.setdefault("autocompletion", autocompletion)
     return click.option(*param_decls, **option_attrs)
 
 
 def argument(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
-    arg_attrs = attrs.copy()
-    typ = convert_type(attrs.get("type"), attrs.get("default"))
-    autocompletion = getattr(typ, "complete", None)
-    arg_attrs.setdefault("autocompletion", autocompletion)
-    return click.argument(*param_decls, **arg_attrs)
+    return click.argument(*param_decls, **attrs)
 
 
 def volume_to_verbose_str(volume: Volume) -> str:
@@ -497,6 +488,40 @@ async def resolve_disk(
     return disk.id
 
 
+BUCKET_ID_PATTERN = (
+    r"bucket-[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}"
+)
+
+
+async def resolve_bucket(
+    id_or_name: str, *, client: Client, cluster_name: Optional[str] = None
+) -> str:
+    # Temporary fast path.
+    if re.fullmatch(BUCKET_ID_PATTERN, id_or_name):
+        return id_or_name
+
+    bucket = await client.buckets.get(id_or_name, cluster_name)
+    return bucket.id
+
+
+BUCKET_CREDENTIAL_ID_PATTERN = (
+    r"bucket-credentials-[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}"
+)
+
+
+async def resolve_bucket_credential(
+    id_or_name: str, *, client: Client, cluster_name: Optional[str] = None
+) -> str:
+    # Temporary fast path.
+    if re.fullmatch(BUCKET_CREDENTIAL_ID_PATTERN, id_or_name):
+        return id_or_name
+
+    credential = await client.buckets.persistent_credentials_get(
+        id_or_name, cluster_name
+    )
+    return credential.id
+
+
 SHARE_SCHEMES = ("storage", "image", "job", "blob", "role", "secret", "disk")
 
 
@@ -532,14 +557,12 @@ def parse_file_resource(uri: str, root: Root) -> URL:
 
 
 def parse_blob_resource(uri: str, root: Root) -> URL:
-    # Username will not be used, just part of the signature
     return uri_from_cli(
         uri, root.client.username, root.client.cluster_name, allowed_schemes=("blob",)
     )
 
 
 def parse_blob_or_file_resource(uri: str, root: Root) -> URL:
-    # Username will not be used, just part of the signature
     return uri_from_cli(
         uri,
         root.client.username,
@@ -567,41 +590,6 @@ def parse_permission_action(action: str) -> Action:
         )
 
 
-def do_deprecated_quiet(
-    ctx: click.Context, param: Union[click.Option, click.Parameter], value: Any
-) -> Any:
-    if ctx.obj is None:
-        return
-    if value and not ctx.obj.quiet:
-        ctx.obj.verbosity = -2
-        click.echo(
-            click.style(
-                "DeprecationWarning: "
-                "The local option --quiet is deprecated. "
-                "Use global option --quiet instead.",
-                fg="red",
-            ),
-            err=True,
-        )
-        # Patch the logger as it was set up with verbosity=-2.
-        root_logger = logging.getLogger()
-        handler = root_logger.handlers[-1]
-        assert handler.formatter
-        format_class = type(handler.formatter)
-        handler.setFormatter(format_class())
-        handler.setLevel(logging.ERROR)
-
-
-deprecated_quiet_option: Any = option(
-    "-q",
-    "--quiet",
-    is_flag=True,
-    callback=do_deprecated_quiet,
-    help="Run command in quiet mode (DEPRECATED)",
-    expose_value=False,
-    is_eager=True,
-)
-
 if sys.version_info >= (3, 7):  # pragma: no cover
     from contextlib import AsyncExitStack
 else:
