@@ -149,11 +149,8 @@ def test_e2e_blob_storage_copy_recursive_folder(
     srcfile, checksum, dir_path = nested_data
     target_file_name = Path(srcfile).name
 
-    # Upload local folder .../neested_data/nested to bucket root
-    captured = helper.run_cli(["blob", "cp", "-r", dir_path, f"blob:{tmp_bucket}"])
-    # stderr has logs like "Using path ..."
-    # assert not captured.err
-    assert not captured.out
+    # Upload local folder .../nested_data/nested to bucket root
+    helper.run_cli(["blob", "cp", "-r", dir_path, f"blob:{tmp_bucket}"])
 
     helper.check_blob_size(
         tmp_bucket, f"nested/directory/for/test/{target_file_name}", FILE_SIZE_B // 3
@@ -211,12 +208,10 @@ def test_e2e_blob_storage_glob_copy(
     ]
 
     # Test subcommand "glob"
-    captured = helper.run_cli(["blob", "glob", f"blob:{tmp_bucket}/folder/*"])
+    captured = helper.run_cli(["blob", "glob", f"blob:{tmp_bucket}/folder/*o*"])
     assert sorted(captured.out.splitlines()) == [
-        prefix,
-        prefix + "bar",
-        prefix + "baz",
         prefix + "foo",
+        prefix + "subfolder/",
     ]
 
     # Download files with pattern
@@ -285,3 +280,37 @@ def test_e2e_blob_storage_cp_filter(
         ]
     )
     assert os.listdir(tmp_path / "filtered") == ["bar"]
+
+
+@pytest.mark.e2e
+def test_e2e_blob_storage_rm_file(
+    helper: Helper, nested_data: Tuple[str, str, str], tmp_path: Path, tmp_bucket: str
+) -> None:
+    srcfile = tmp_path / "testfile"
+    srcfile.write_bytes(b"abc")
+
+    captured = helper.run_cli(["blob", "cp", str(srcfile), f"blob:{tmp_bucket}"])
+    assert not captured.out
+
+    helper.run_cli(["blob", "rm", f"blob:{tmp_bucket}/testfile"])
+
+    captured = helper.run_cli(["blob", "ls", f"blob:{tmp_bucket}/testfile"])
+    assert not captured.out
+
+
+@pytest.mark.e2e
+def test_e2e_blob_storage_rm_dir(
+    helper: Helper, nested_data: Tuple[str, str, str], tmp_path: Path, tmp_bucket: str
+) -> None:
+    srcfile = tmp_path / "testfile"
+    srcfile.write_bytes(b"abc")
+
+    captured = helper.run_cli(
+        ["blob", "cp", str(srcfile), f"blob:{tmp_bucket}/some_dir/"]
+    )
+    assert not captured.out
+
+    helper.run_cli(["blob", "rm", "-r", f"blob:{tmp_bucket}/some_dir/"])
+
+    captured = helper.run_cli(["blob", "ls", f"blob:{tmp_bucket}/some_dir/"])
+    assert not captured.out
